from cobald.interfaces import Pool, PoolDecorator


class Coarser(PoolDecorator):
    """
    Proxy limiting changes in a pool to a certain granularity

    :param target: the pool to which changes are applied
    :param granularity: granularity of changes
    """
    @property
    def demand(self):
        return self._demand

    @demand.setter
    def demand(self, value):
        self._demand = value
        self.target.demand = value // self.granularity * self.granularity

    def __init__(self, target: Pool, granularity: int = 1):
        super().__init__(target=target)
        self._demand = 0
        self.granularity = granularity



from collections.abc import Mapping
from logging import Formatter, LogRecord
import json


#: Attributes of a LogRecord. See <the docs `https://docs.python.org/3/library/logging.html#logrecord-attributes`>_.
RECORD_ATTRIBUTES = (
    'args', 'asctime', 'created', 'exc_info', 'exc_text', 'filename', 'funcName',
    'levelname', 'levelno', 'lineno', 'message', 'module', 'msecs',
    'msg', 'name', 'pathname', 'process', 'processName', 'relativeCreated',
    'stack_info', 'thread', 'threadName',
)


class JsonFormatter(Formatter):
    """
    Formatter that emits data as JSON

    :param fmt: default data for all records
    :param datefmt: format for timestamps

    The ``datefmt`` parameter has almost the same meaning as for :py:class:`~.Formatter`.
    Setting it to ``None`` uses the default time format.
    However, setting it to any other value that is boolean false excludes the timestamp from reports.
    """
    def __init__(self, fmt: dict = None, datefmt: str = None):
        super().__init__(fmt=None, datefmt=datefmt, style='%')
        self._defaults = fmt or {}
        if not isinstance(self._defaults, Mapping):
            raise TypeError('`fmt` must be a Mapping or None')
        self._add_time = self.datefmt or self.datefmt is None

    def format(self, record: LogRecord):
        args = record.args
        assert isinstance(args, Mapping), 'monitor record argument must be a mapping, not %r' % type(args)
        data = self._defaults
        if self._add_time:
            data['time'] = self.formatTime(record, self.datefmt)
        data['message'] = record.getMessage()
        data.update(record.args)
        return json.dumps(data)


if __name__ == '__main__':
    import logging
    logger = logging.getLogger()
    logger.handlers = [logging.StreamHandler()]
    logger.handlers[0].formatter = JsonFormatter({'latitude': 49, 'longitude': 8})
    logger.warning('forecast', {'temperature': 298, 'humidity': 0.45})



import threading
import pytest
import time

import logging

from cobald.daemon.service import ServiceRunner, service


logging.getLogger().level = 10


class TerminateRunner(Exception):
    pass


def run_in_thread(payload, name, daemon=True):
    thread = threading.Thread(target=payload, name=name, daemon=daemon)
    thread.start()
    time.sleep(0.0)


class TestServiceRunner(object):
    def test_no_tainting(self):
        """Assert that no payloads may be scheduler before starting"""
        def payload():
            return

        runner = ServiceRunner()
        runner._meta_runner.register_payload(payload, flavour=threading)
        with pytest.raises(RuntimeError):
            runner.accept()

    def test_service(self):
        runner = ServiceRunner(accept_delay=0.1)
        replies = []

        @service(flavour=threading)
        class Service(object):
            def __init__(self):
                self.done = threading.Event()
                self.done.clear()

            def run(self):
                replies.append(1)
                self.done.set()

        a = Service()
        run_in_thread(runner.accept, name='test_service')
        assert a.done.wait(timeout=5), 'service thread completed'
        assert len(replies) == 1, 'pre-registered service ran'
        b = Service()
        assert b.done.wait(timeout=5), 'service thread completed'
        assert len(replies) == 2, 'post-registered service ran'
        runner.shutdown()



import logging
import weakref
import trio
import gc
import functools
import threading

from types import ModuleType

from ..utility.concurrent.meta_runner import MetaRunner


class ServiceUnit(object):
    __active_units__ = weakref.WeakSet()

    def __init__(self, service, flavour):
        assert hasattr(service, 'run'), "service must implement a 'run' method"
        assert any(flavour == runner.flavour for runner in MetaRunner.runner_types), \
            "service flavour must be one of %s" % ','.join(repr(runner.flavour) for runner in MetaRunner.runner_types)
        self.service = weakref.ref(service)
        self.flavour = flavour
        self._started = False
        ServiceUnit.__active_units__.add(self)

    @classmethod
    def units(cls):
        """Sequence of all currently defined units"""
        return list(cls.__active_units__)

    @property
    def running(self):
        return self._started

    def start(self, runner: MetaRunner):
        service = self.service()
        if service is None:
            return
        else:
            self._started = True
            runner.register_payload(service.run, flavour=self.flavour)

    def __repr__(self):
        return '%s(%r, flavour=%r)' % (self.__class__.__name__, self.service() or '<defunct>', self.flavour)


def service(flavour):
    """
    Mark a class as implementing a Service

    Each Service class must have a ``run`` method.
    This method is :py:meth:`~.ServiceRunner.adopt`\ ed after the daemon starts, unless

    * the Service has been garbage collected, or
    * the ServiceUnit has been :py:meth:`~.ServiceUnit.cancel`\ ed.

    For each service instance, its :py:class:`~.ServiceUnit` is available at ``service_instance.__service_unit__``.
    """
    def service_unit_decorator(raw_cls):
        __new__ = raw_cls.__new__

        def __new_service__(cls, *args, **kwargs):
            if __new__ is object.__new__:
                self = __new__(cls)
            else:
                self = __new__(cls, *args, **kwargs)
            service_unit = ServiceUnit(self, flavour)
            self.__service_unit__ = service_unit
            return self

        raw_cls.__new__ = __new_service__
        if raw_cls.run.__doc__ is None:
            raw_cls.run.__doc__ = "Service entry point"
        return raw_cls
    return service_unit_decorator


class ServiceRunner(object):
    """
    Runner for coroutines, subroutines and services
    """
    def __init__(self, accept_delay: float = 1):
        self._logger = logging.getLogger('cobald.runtime.daemon.services')
        self._meta_runner = MetaRunner()
        self._must_shutdown = False
        self._is_shutdown = threading.Event()
        self.accept_delay = accept_delay

    def execute(self, payload, flavour: ModuleType):
        """Synchronously run ``payload`` and provide its output"""
        self._meta_runner.run_payload(payload, flavour=flavour)

    def adopt(self, payload, flavour: ModuleType):
        """Concurrently run ``payload`` in the background"""
        self._meta_runner.register_payload(payload, flavour=flavour)

    def accept(self):
        """Start accepting synchronous, asynchronous and service payloads"""
        if self._meta_runner:
            raise RuntimeError('payloads scheduled for %s before being started' % self)
        self._must_shutdown = False
        self._logger.info('ServiceRunner starting')
        # force collecting objects so that defunct, migrated and overwritten services are destroyed now
        gc.collect()
        self._adopt_services()
        self.adopt(self.run, flavour=trio)
        self._meta_runner.run()

    def shutdown(self):
        """Shutdown the accept loop and stop running payloads"""
        self._must_shutdown = True
        self._is_shutdown.wait()
        self._meta_runner.stop()

    async def run(self):
        delay, max_delay, increase = 0.0, self.accept_delay, self.accept_delay / 10
        self._is_shutdown.clear()
        while not self._must_shutdown:
            self._adopt_services()
            await trio.sleep(delay)
            delay = min(delay + increase, max_delay)
        self._is_shutdown.set()

    def _adopt_services(self):
        for unit in ServiceUnit.units():  # type: ServiceUnit
            if unit.running:
                continue
            unit.start(self._meta_runner)



import argparse

CLI = argparse.ArgumentParser(description='COBalD - the Opportunistic Balancing Daemon')
CLI.add_argument(
    'CONFIGURATION',
    help='path of the configuration to use',
    type=str,
)
CLI_LOG = CLI.add_argument_group('Startup Logging')
CLI_LOG.add_argument(
    '--log-level',
    help='initial logging level',
    default='INFO',
    choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
    type=str.upper,
)
CLI_LOG.add_argument(
    '--log-target',
    help='initial logging target; stdout, stderr or a file path',
    default='stderr',
)
CLI_LOG.add_argument(
    '--log-journal',
    help='use short formatting suitable for journals',
    action='store_true',
)



from ._pool import Pool


class PoolDecorator(Pool):
    """
    Decorator modifying how a pool provides resources
    """
    def __init__(self, target: Pool):
        self.target = target

    @property
    def supply(self):
        """The volume of resources that is provided by this site"""
        return self.target.supply

    @property
    def demand(self):
        """The volume of resources to be provided by this site"""
        return self.target.demand

    @demand.setter
    def demand(self, value):
        self.target.demand = value

    @property
    def utilisation(self) -> float:
        """Fraction of the provided resources which is actively used"""
        return self.target.utilisation

    @property
    def allocation(self) -> float:
        """Fraction of the provided resources which is assigned for usage"""
        return self.target.allocation



import trio
from functools import partial


from .base_runner import BaseRunner
from .async_tools import raise_return, AsyncExecution


class TrioRunner(BaseRunner):
    """Runner for coroutines of :py:mod:`trio`"""
    flavour = trio

    def __init__(self):
        self._nursery = None
        super().__init__()

    def register_payload(self, payload):
        super().register_payload(partial(raise_return, payload))

    def run_payload(self, payload):
        execution = AsyncExecution(payload)
        super().register_payload(execution.coroutine)
        return execution.wait()

    def _run(self):
        return trio.run(self.await_all)

    async def await_all(self):
        delay = 0.0
        async with trio.open_nursery() as nursery:
            while self.running.is_set():
                await self._start_outstanding(nursery=nursery)
                await trio.sleep(delay)
                delay = min(delay + 0.1, 1.0)
            nursery.cancel_scope.cancel()

    async def _start_outstanding(self, nursery):
        with self._lock:
            for coroutine in self._payloads:
                nursery.start_soon(coroutine)
            self._payloads.clear()
        await trio.sleep(0)



import asyncio
from functools import partial

from .base_runner import BaseRunner
from .async_tools import raise_return, AsyncExecution


class AsyncioRunner(BaseRunner):
    flavour = asyncio

    def __init__(self):
        super().__init__()
        self.event_loop = asyncio.new_event_loop()
        self._tasks = set()

    def register_payload(self, payload):
        super().register_payload(partial(raise_return, payload))

    def run_payload(self, payload):
        execution = AsyncExecution(payload)
        super().register_payload(execution.coroutine)
        return execution.wait()

    def _run(self):
        asyncio.set_event_loop(self.event_loop)
        self.event_loop.run_until_complete(self.await_all())

    async def await_all(self):
        delay = 0.0
        try:
            while self.running.is_set():
                await self._start_outstanding()
                await self._manage_running()
                await asyncio.sleep(delay)
                delay = min(delay + 0.1, 1.0)
        except Exception:
            await self._cancel_running()
            raise

    async def _start_outstanding(self):
        with self._lock:
            for coroutine in self._payloads:
                task = self.event_loop.create_task(coroutine())
                self._tasks.add(task)
            self._payloads.clear()
        await asyncio.sleep(0)

    async def _manage_running(self):
        for task in self._tasks.copy():
            if task.done():
                self._tasks.remove(task)
                if task.exception() is not None:
                    raise task.exception()
        await asyncio.sleep(0)

    async def _cancel_running(self):
        for task in self._tasks:
            task.cancel()
            await asyncio.sleep(0)
        for task in self._tasks:
            while not task.done():
                await asyncio.sleep(0.1)

    def stop(self):
        if not self.running.wait(0.2):
            return
        self._logger.debug('runner disabled: %s', self)
        with self._lock:
            self.running.clear()
            for task in self._tasks:
                task.cancel()
        self._stopped.wait()
        self.event_loop.stop()
        self.event_loop.close()



"""
Interfaces for primitives of the :py:mod:`cobald` model

Each :py:class:`~.Pool` provides a varying number of resources.
A :py:class:`~.Controller` adjusts the number of resources that a :py:class:`~.Pool` must provide.
Several :py:class:`~.Pool` can be combined in a single :py:class:`~.CompositePool` to appear as one.
To modify how a :py:class:`~.Pool` presents or digests data,
any number of :py:class:`~.PoolDecorator` may proceed it.

.. graphviz::

    digraph graphname {
        graph [rankdir=LR, splines=lines, bgcolor="transparent"]
        controller [label=Controller]
        composite [label=CompositePool]
        decoa [label=PoolDecorator]
        poola, poolb [label=Pool]
        controller -> decoa -> composite
        composite -> poola
        composite -> poolb
    }

"""
from ._composite import CompositePool
from ._controller import Controller
from ._pool import Pool
from ._proxy import PoolDecorator

__all__ = [cls.__name__ for cls in (Pool, PoolDecorator, Controller, CompositePool)]



import threading
import time

from ..debug import NameRepr
from .base_runner import BaseRunner, OrphanedReturn


class CapturingThread(threading.Thread):
    """
    Daemonic threads that automatically capture any return value or exception from their ``target``
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs, daemon=True)
        self._exception = None
        self._name = str(NameRepr(self._target))

    def join(self, timeout=None):
        super().join(timeout=timeout)
        if self._started.is_set() and not self.is_alive():
            if self._exception is not None:
                raise self._exception
        return not self.is_alive()

    def run(self):
        """Modified ``run`` that captures return value and exceptions from ``target``"""
        try:
            if self._target:
                return_value = self._target(*self._args, **self._kwargs)
                if return_value is not None:
                    self._exception = OrphanedReturn(self, return_value)
        except BaseException as err:
            self._exception = err
        finally:
            # Avoid a refcycle if the thread is running a function with
            # an argument that has a member that points to the thread.
            del self._target, self._args, self._kwargs


class ThreadRunner(BaseRunner):
    """Runner for subroutines with :py:mod:`threading`"""
    flavour = threading

    def __init__(self):
        super().__init__()
        self._threads = set()

    def run_payload(self, payload):
        # - run_payload has to block until payload is done
        # instead of running payload in a thread and blocking this one,
        # we just block this thread by running payload directly
        return payload()

    def _run(self):
        delay = 0.0
        while self.running.is_set():
            self._start_outstanding()
            for thread in self._threads.copy():
                if thread.join(timeout=0):
                    self._threads.remove(thread)
                    self._logger.debug('reaped thread %s', thread)
            time.sleep(delay)
            delay = min(delay + 0.1, 1.0)

    def _start_outstanding(self):
        with self._lock:
            payloads = self._payloads.copy()
            self._payloads.clear()
        for subroutine in payloads:
            thread = CapturingThread(target=subroutine)
            thread.start()
            self._threads.add(thread)
            self._logger.debug('booted thread %s', thread)
        time.sleep(0)



import logging

from cobald.interfaces import Pool, PoolDecorator


_DEFAULT_MESSAGE = \
    'demand = %(value)s '\
    '[demand=%(demand)s, supply=%(supply)s, utilisation=%(utilisation).2f, consumption=%(consumption).2f]'


class Logger(PoolDecorator):
    """
    Proxy which logs all changes of ``target.demand``

    :param name: name of the underlying :py:class:`logging.Logger`
    :param message: message to emit for every change
    :param level: numerical logging level

    The ``message`` parameter is used as a ``%``-style format string with named fields.
    Valid named fields are

    ``value``
        for the new demand to set,

    ``demand``, ``supply``, ``utilisation`` and ``consumption``
        for the current state of ``target``, and

    ``target``
        for the raw ``target`` pool.
    """
    @property
    def demand(self):
        return self.target.demand

    @demand.setter
    def demand(self, value):
        self._logger.log(
            self.level, self.message,
            {
                'value': value, 'demand': self.target.demand, 'supply': self.target.supply,
                'utilisation': self.target.utilisation, 'consumption': self.target.allocation,
                'target': self.target,
            })
        self.target.demand = value

    @property
    def name(self) -> str:
        return self._logger.name

    @name.setter
    def name(self, value: str):
        if value is None:
            value = self.target.__class__.__qualname__
        self._logger = logging.getLogger(value)

    def __init__(self, target: Pool, name: str = None, message: str = _DEFAULT_MESSAGE, level: int = logging.INFO):
        super().__init__(target=target)
        self._logger = None  # type: logging.Logger
        self.message = message
        self.name = name
        self.level = level



from cobald.interfaces import Pool


class MockPool(Pool):
    demand = 0

    def __init__(self):
        self._utilisation = 1.0
        self._allocation = 1.0

    @property
    def supply(self):
        return self.demand

    @property
    def allocation(self):
        return self._allocation

    @allocation.setter
    def allocation(self, value):
        if value < self._utilisation:
            self._utilisation = value
        self._allocation = value

    @property
    def utilisation(self):
        return self._utilisation

    @utilisation.setter
    def utilisation(self, value):
        if value > self._allocation:
            self._allocation = value
        self._utilisation = value



import threading
import pytest
import time
import asyncio

import trio

from cobald.utility.concurrent.base_runner import OrphanedReturn
from cobald.utility.concurrent.meta_runner import MetaRunner


class TerminateRunner(Exception):
    pass


def run_in_thread(payload, name, daemon=True):
    thread = threading.Thread(target=payload, name=name, daemon=daemon)
    thread.start()
    time.sleep(0.0)


class TestMetaRunner(object):
    def test_bool_payloads(self):
        def subroutine():
            time.sleep(0.5)

        async def a_coroutine():
            await asyncio.sleep(0.5)

        async def t_coroutine():
            await trio.sleep(0.5)

        for flavour, payload in ((threading, subroutine), (asyncio, a_coroutine), (trio, t_coroutine)):
            runner = MetaRunner()
            assert not bool(runner)
            runner.register_payload(payload, flavour=flavour)
            assert bool(runner)
            run_in_thread(runner.run, name='test_bool_payloads %s' % flavour)
            assert bool(runner)
            runner.stop()

    def test_run_subroutine(self):
        """Test executing a subroutine"""
        def with_return():
            return 'expected return value'

        def with_raise():
            raise KeyError('expected exception')

        for flavour in (threading,):
            runner = MetaRunner()
            result = runner.run_payload(with_return, flavour=flavour)
            assert result == with_return()
            with pytest.raises(KeyError):
                runner.run_payload(with_raise, flavour=flavour)

    def test_run_coroutine(self):
        """Test executing a subroutine"""
        async def with_return():
            return 'expected return value'

        async def with_raise():
            raise KeyError('expected exception')

        for flavour in (trio, asyncio):
            runner = MetaRunner()
            run_in_thread(runner.run, name='test_run_coroutine %s' % flavour)
            result = runner.run_payload(with_return, flavour=flavour)
            # TODO: can we actually get the value from with_return?
            assert result == 'expected return value'
            with pytest.raises(KeyError):
                runner.run_payload(with_raise, flavour=flavour)
            runner.stop()

    def test_return_subroutine(self):
        """Test that returning from subroutines aborts runners"""
        def with_return():
            return 'unhandled return value'

        for flavour in (threading,):
            runner = MetaRunner()
            runner.register_payload(with_return, flavour=flavour)
            with pytest.raises(RuntimeError) as exc:
                runner.run()
            assert isinstance(exc.value.__cause__, OrphanedReturn)

    def test_return_coroutine(self):
        """Test that returning from subroutines aborts runners"""
        async def with_return():
            return 'unhandled return value'

        for flavour in (asyncio, trio):
            runner = MetaRunner()
            runner.register_payload(with_return, flavour=flavour)
            with pytest.raises(RuntimeError) as exc:
                runner.run()
            assert isinstance(exc.value.__cause__, OrphanedReturn)

    def test_abort_subroutine(self):
        """Test that failing subroutines abort runners"""
        def abort():
            raise TerminateRunner

        for flavour in (threading,):
            runner = MetaRunner()
            runner.register_payload(abort, flavour=flavour)
            with pytest.raises(RuntimeError) as exc:
                runner.run()
            assert isinstance(exc.value.__cause__, TerminateRunner)

            def noop():
                return

            def loop():
                while True:
                    time.sleep(0)

            runner = MetaRunner()
            runner.register_payload(noop, loop, flavour=flavour)
            runner.register_payload(abort, flavour=flavour)
            with pytest.raises(RuntimeError) as exc:
                runner.run()
            assert isinstance(exc.value.__cause__, TerminateRunner)

    def test_abort_coroutine(self):
        """Test that failing coroutines abort runners"""
        async def abort():
            raise TerminateRunner

        for flavour in (asyncio, trio):
            runner = MetaRunner()
            runner.register_payload(abort, flavour=flavour)
            with pytest.raises(RuntimeError) as exc:
                runner.run()
            assert isinstance(exc.value.__cause__, TerminateRunner)

            async def noop():
                return

            async def loop():
                while True:
                    await flavour.sleep(0)
            runner = MetaRunner()

            runner.register_payload(noop, loop, flavour=flavour)
            runner.register_payload(abort, flavour=flavour)
            with pytest.raises(RuntimeError) as exc:
                runner.run()
            assert isinstance(exc.value.__cause__, TerminateRunner)



from yaml import load

from .mapping import configure_logging, Translator, ConfigurationError


def load_configuration(path, translator=Translator()):
    with open(path) as yaml_stream:
        config_data = load(yaml_stream)
    try:
        logging_mapping = config_data.pop('logging')
    except KeyError:
        pass
    else:
        configure_logging(logging_mapping)
    try:
        root_pipeline = config_data.pop('pipeline')
    except KeyError as err:
        raise ConfigurationError(where='root', what=err)
    else:
        if config_data:
            raise ConfigurationError(where='root', what='dangling configuration keys (%s)' % ', '.join(config_data))
        return translator.translate_hierarchy({'pipeline': root_pipeline})



class NameRepr(object):
    """
    Lazy pretty formatter for name of objects
    """
    def __init__(self, target):
        self.target = target

    def __str__(self):
        target = self.target
        try:
            return target.__module__ + ':' + target.__qualname__
        except AttributeError:
            try:
                return target.__module__ + ':' + target.__name__
            except AttributeError:
                return target.__name__

    __repr__ = __str__



"""
===========================================
COBalD - the Opportunistic Balancing Daemon
===========================================

.. image:: https://readthedocs.org/projects/cobald/badge/?version=latest
    :target: http://cobald.readthedocs.io/en/latest/?badge=latest
    :alt: Documentation Status

This is a **draft** for a feedback based balancing system for providing opportunistic resources.
"""
__title__ = 'cobald'
__summary__ = 'COBalD - the Opportunistic Balancing Deamon'
__url__ = 'https://github.com/MaineKuehn/cobald'

__version__ = '0.9.0'
__author__ = 'Eileen Kuehn, Max Fischer'
__email__ = 'mainekuehn@gmail.com'
__copyright__ = '2018 %s' % __author__
__keywords__ = 'opportunistic scheduling scheduler demand feedback-loop cobald'



import logging
import threading
import trio

from types import ModuleType

from .base_runner import BaseRunner
from .trio_runner import TrioRunner
from .asyncio_runner import AsyncioRunner
from .thread_runner import ThreadRunner


from ...utility.debug import NameRepr


class MetaRunner(object):
    """
    Unified interface to schedule subroutines and coroutines for concurrent execution
    """
    runner_types = (TrioRunner, AsyncioRunner, ThreadRunner)

    def __init__(self):
        self._logger = logging.getLogger('cobald.runtime.runner.meta')
        self.runners = {
            runner.flavour: runner() for runner in self.runner_types
        }  # type: dict[ModuleType, BaseRunner]
        self._lock = threading.Lock()
        self.running = threading.Event()
        self.running.clear()

    def __bool__(self):
        return any(bool(runner) for runner in self.runners.values())

    def register_payload(self, *payloads, flavour: ModuleType):
        """Queue one or more payload for execution after its runner is started"""
        for payload in payloads:
            self._logger.debug('registering payload %s (%s)', NameRepr(payload), NameRepr(flavour))
            self.runners[flavour].register_payload(payload)

    def run_payload(self, payload, *, flavour: ModuleType):
        """Execute one payload after its runner is started and return its output"""
        return self.runners[flavour].run_payload(payload)

    def run(self):
        """Run all runners until completion"""
        self._logger.info('starting all runners')
        try:
            with self._lock:
                assert not self.running.set(), 'cannot re-run: %s' % self
                self.running.set()
            thread_runner = self.runners[threading]
            for runner in self.runners.values():
                if runner is not thread_runner:
                    thread_runner.register_payload(runner.run)
            thread_runner.run()
        except Exception as err:
            self._logger.error('runner terminated: %s', err)
            raise RuntimeError from err
        finally:
            self._stop_runners()
            self._logger.info('stopped all runners')
            self.running.clear()

    def stop(self):
        """Stop all runners"""
        self._stop_runners()

    def _stop_runners(self):
        for runner in self.runners.values():
            if runner.flavour == threading:
                continue
            runner.stop()
        self.runners[threading].stop()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    import time
    import asyncio
    runner = MetaRunner()

    async def trio_sleeper():
        for i in range(3):
            print('trio\t', i)
            await trio.sleep(0.1)
    runner.register_payload(trio_sleeper, flavour=trio)

    async def asyncio_sleeper():
        for i in range(3):
            print('asyncio\t', i)
            await asyncio.sleep(0.1)
    runner.register_payload(asyncio_sleeper, flavour=asyncio)

    def thread_sleeper():
        for i in range(3):
            print('thread\t', i)
            time.sleep(0.1)
    runner.register_payload(thread_sleeper, flavour=threading)

    async def teardown():
        await trio.sleep(5)
        raise SystemExit('Abort from trio runner')
    runner.register_payload(teardown, flavour=trio)

    runner.run()



import trio

from cobald.interfaces import Pool, PoolDecorator

from cobald.daemon import service


@service(flavour=trio)
class Buffer(PoolDecorator):
    """
    A timed buffer for changes to a pool

    :param target: the pool to which changes are applied
    :param window: interval after which changes are applied

    Any changes made to :py:attr:`demand` are stored internally.
    Every ``window`` seconds, the final demand is applied to ``target``.
    """
    demand = 0.0

    def __init__(self, target: Pool, window: float = 10.):
        super().__init__(target=target)
        self.window = window
        self.demand = target.demand

    async def run(self):
        while True:
            if self.demand != self.target.demand:
                self.target.demand = self.demand
            await trio.sleep(self.window)



import json
import io
import logging
import threading
import time


from cobald.monitor.format_json import JsonFormatter


class CapturingHandler(logging.StreamHandler):
    @property
    def content(self):
        return self.stream.getvalue()

    def __init__(self):
        super().__init__(stream=io.StringIO())


_test_index = 0
_index_lock = threading.Lock()


class ExtraLogger(logging.Logger):
    def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
        """Replacement for Logger.makeRecord to overwrite fields via ``extra``"""
        try:
            created = extra and extra.pop('created')
        except KeyError:
            created = None
        rv = super().makeRecord(name, level, fn, lno, msg, args, exc_info, func, None, sinfo)
        if extra is not None:
            for key in extra:
                rv.__dict__[key] = extra[key]
        if created:
            rv.created = created
            rv.msecs = (created - int(created)) * 1000
            rv.relativeCreated = (created - logging._startTime) * 1000
        return rv


def make_test_logger(base_name: str = 'test_logger'):
    with _index_lock:
        global _test_index
        log_name = base_name + '.test%d' % _test_index
        _test_index += 1
    logger = logging.getLogger(log_name)
    logger.__class__ = ExtraLogger
    handler = CapturingHandler()
    logger.handlers = [handler]
    return logger, handler


class TestFormatJson(object):
    def test_payload(self):
        for payload in (
                {'a': 'a'},
                {str(i): i for i in range(20)},
        ):
            logger, handler = make_test_logger(__name__)
            handler.formatter = JsonFormatter()
            logger.critical('message', payload)
            data = json.loads(handler.content)
            assert len(data) == len(payload) + 2
            assert data.pop('message') == 'message'
            assert data.pop('time')
            assert data == payload

    def test_default_timestamp(self):
        now = time.time()
        payload = {'a': 'a', '1': 1, '2.2': 2.2}
        logger, handler = make_test_logger(__name__)
        handler.formatter = JsonFormatter()
        logger.critical('message', payload, extra={'created': now})
        data = json.loads(handler.content)
        assert len(data) == len(payload) + 2
        # from Formatter and LogRecord
        ct = time.localtime(now)
        msecs = (now - int(now)) * 1000
        default_time_string = logging.Formatter.default_msec_format % (
            time.strftime(logging.Formatter.default_time_format, ct),
            msecs
        )
        assert data.pop('time') == default_time_string

    def test_explicit_timestamp(self):
        now = time.time()
        payload = {'a': 'a', '1': 1, '2.2': 2.2}
        logger, handler = make_test_logger(__name__)
        handler.formatter = JsonFormatter(datefmt='%Y')
        logger.critical('message', payload, extra={'created': now})
        data = json.loads(handler.content)
        assert len(data) == len(payload) + 2
        # from Formatter and LogRecord
        ct = time.localtime(now)
        default_time_string = time.strftime('%Y', ct)
        assert data.pop('time') == default_time_string

    def test_disabled_timestamp(self):
        now = time.time()
        payload = {'a': 'a', '1': 1, '2.2': 2.2}
        logger, handler = make_test_logger(__name__)
        handler.formatter = JsonFormatter(datefmt='')
        logger.critical('message', payload, extra={'created': now})
        data = json.loads(handler.content)
        assert len(data) == len(payload) + 1
        assert 'time' not in data




import sys
import logging
import logging.handlers


def create_handler(target: str):
    """Create a handler for logging to ``target``"""
    if target == 'stderr':
        return logging.StreamHandler(sys.stderr)
    elif target == 'stdout':
        return logging.StreamHandler(sys.stdout)
    else:
        return logging.handlers.WatchedFileHandler(filename=target)


def initialise_logging(level: str, target: str, short_format: bool):
    """Initialise basic logging facilities"""
    try:
        log_level = getattr(logging, level)
    except AttributeError:
        raise SystemExit(
            "invalid log level %r, expected any of 'DEBUG', 'INFO', 'WARNING', 'ERROR' or 'CRITICAL'" % level
        )
    handler = create_handler(target=target)
    logging.basicConfig(
        level=log_level,
        format='%(asctime)-15s (%(process)d) %(message)s' if not short_format else '%(message)s',
        datefmt='%Y-%m-%d %H:%M:%S',
        handlers=[handler]
    )



import abc
import asyncio
import warnings


class Actor(metaclass=abc.ABCMeta):
    """
    An active component that can be run by an event loop
    """
    def __new__(cls, *args, **kwargs):
        warnings.warn(FutureWarning(
            'Actor will be removed in the future. '
            '%s Instances should autonomously register themselves with `deamon.runner`.' % cls.__name__
        ))
        return super().__new__(cls)

    def mount(self, event_loop: asyncio.AbstractEventLoop):
        """Mount the Actor in ``event_loop`` for execution"""
        event_loop.create_task(self.run())

    @asyncio.coroutine
    def run(self):
        return



import abc

from ._pool import Pool


class Controller(metaclass=abc.ABCMeta):
    """
    Controller adjusting the demand in a :py:class:`~.Pool`

    :param target: the resource pool for which demand is adjusted
    """
    def __init__(self, target: Pool):
        self.target = target



from typing import Callable
import weakref

import trio

from cobald.interfaces import Pool, CompositePool
from cobald.daemon import service


@service(flavour=trio)
class FactoryPool(CompositePool):
    """
    Composition that adds and removes pools to satisfy demand

    :param factory: a callable that produces a single new :py:class:`~.Pool` on each invocation
    :param interval: how often to adjust the number of children

    Adjustment uses two extensions that children must respond to adequately:

    * When spawned via ``factory()``, children shall already be set to their expected ``demand``.
    * When disabled via ``demand=0``, children shall shut down and free any resources and tasks.

    Once spawned, children are free to adjust their demand if required.
    The :py:class:`FactoryPool` re-inspects demand before spawning or disabling children.
    """
    @property
    def children(self):
        return [*self._hatchery, *self._mortuary]

    @property
    def demand(self):
        return self._demand

    @demand.setter
    def demand(self, value):
        # we may spend an arbitrary time spawning Drones, just acknowledge demand and defer any actions
        self._demand = value

    @property
    def supply(self):
        return sum(child.supply for child in self.children)

    @property
    def utilisation(self):
        try:
            return sum(child.utilisation for child in self._children) / len(self._children)
        except ZeroDivisionError:
            return 1.

    @property
    def allocation(self):
        try:
            return sum(child.allocation for child in self._children) / len(self._children)
        except ZeroDivisionError:
            return 1.

    def __init__(self, *children: Pool, factory: Callable[..., Pool], interval: float = 30):
        self._demand = sum(child.demand for child in children)
        #: children fulfilling our demand
        self._hatchery = set(children)
        #: children shutting down
        self._mortuary = weakref.WeakSet()
        self.factory = factory
        self.interval = interval

    async def run(self):
        while True:
            await trio.sleep_until(self.interval)
            # freeze target demand in case another thread updates us
            supply, demand = self.supply, self.demand
            if supply > demand:
                await self._shrink(target=demand)
            else:
                await self._grow(target=demand)

    async def _shrink(self, target: float):
        # we can only reap children that are not already shutting down
        # prefer reaping children that supply few used resources
        hit_list = sorted(self._hatchery, key=lambda child: child.supply * child.utilisation)
        excess_demand = sum(child.demand for child in hit_list) - target
        for child in hit_list:
            if excess_demand <= 0:
                break
            # reap child
            if child.demand <= excess_demand:
                excess_demand -= child.demand
                child.demand = 0
                self._hatchery.discard(child)
                self._mortuary.add(child)

    async def _grow(self, target: float):
        missing_demand = target - sum(child.demand for child in self.children)
        while missing_demand > 0:
            new_child = self.factory()
            self._hatchery.add(new_child)
            assert new_child.demand > 0, 'factory must produce children with initial demand'
            missing_demand -= new_child.demand



import logging
import logging.config
import sys

_logger = logging.getLogger(__package__)


class ConfigurationError(Exception):
    def __init__(self, where: str, what):
        self.where = where
        self.what = what
        super().__init__("invalid configuration element '%s': %s" % (where, what))


def configure_logging(logging_mapping):
    _logger.info('Configuring logging')
    logging.config.dictConfig(logging_mapping)


class Translator(object):
    """
    Translator from a mapping to an initialised object hierarchy
    """
    def translate_hierarchy(self, structure, *, where='', **construct_kwargs):
        try:
            if isinstance(structure, dict):
                structure = {
                    key: self.translate_hierarchy(value, where='%s.%s' % (where, key))
                    for key, value in structure.items()
                }
                if '__type__' in structure:
                    return self.construct(structure, **construct_kwargs)
                return structure
            elif isinstance(structure, list):
                # translate bottom up - need those lists to materialize reversed and enumerate iterables
                return list(reversed([
                    self.translate_hierarchy(item, where='%s[%s]' % (where, index))
                    for index, item in reversed(list(enumerate(structure)))
                ]))
            else:
                return structure
        except ConfigurationError:
            raise
        except Exception as err:
            raise ConfigurationError(where=where, what=err)

    def construct(self, mapping: dict, **kwargs):
        """
        Construct an object from a mapping

        :param mapping: the constructor definition, with ``__type__`` name and keyword arguments
        :param kwargs: additional keyword arguments to pass to the constructor
        """
        assert '__type__' not in kwargs and '__args__' not in kwargs
        mapping = {**mapping, **kwargs}
        factory_fqdn = mapping.pop('__type__')
        factory = self.load_name(factory_fqdn)
        args = mapping.pop('__args__', [])
        return factory(*args, **mapping)

    @staticmethod
    def load_name(absolute_name: str):
        """Load an object based on an absolute, dotted name"""
        path = absolute_name.split('.')
        try:
            __import__(absolute_name)
        except ImportError:
            try:
                obj = sys.modules[path[0]]
            except KeyError:
                raise ModuleNotFoundError('No module named %r' % path[0])
            else:
                for component in path[1:]:
                    try:
                        obj = getattr(obj, component)
                    except AttributeError as err:
                        raise ConfigurationError('no such object %r: %s' % (absolute_name, err))
                return obj
        else:  # ImportError is not raised if ``absolute_name`` points to a valid module
            return sys.modules[absolute_name]



from cobald.interfaces import Pool, PoolDecorator

from ..utility.primitives import infinity


class Limiter(PoolDecorator):
    """
    A passive buffer limiting changes to a pool

    :param target: the pool to which changes are applied
    :param minimum: minimum demand to provide
    :param maximum: maximum demand to provide
    """
    @property
    def demand(self):
        return self.target.demand

    @demand.setter
    def demand(self, value):
        self.target.demand = type(value)(min(self.maximum, max(self.minimum, value)))

    def __init__(self, target: Pool, minimum: float = 1.0, maximum: float = infinity):
        super().__init__(target=target)
        self.minimum = minimum
        self.maximum = maximum



import abc


class Pool(metaclass=abc.ABCMeta):
    """
    Individual provider for a number of indistinguishable resources
    """
    @property
    @abc.abstractmethod
    def supply(self):
        """The volume of resources that is provided by this pool"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def demand(self):
        """The volume of resources to be provided by this pool"""
        raise NotImplementedError

    @demand.setter
    @abc.abstractmethod
    def demand(self, value):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def utilisation(self) -> float:
        """Fraction of the provided resources which are actively used"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def allocation(self) -> float:
        """Fraction of the provided resources which are assigned for usage"""
        raise NotImplementedError



import threading
from .base_runner import OrphanedReturn


async def raise_return(payload):
    """Wrapper to raise exception on unhandled return values"""
    value = await payload()
    if value is not None:
        raise OrphanedReturn(payload, value)


class AsyncExecution(object):
    def __init__(self, payload):
        self.payload = payload
        self._result = None
        self._done = threading.Event()
        self._done.clear()

    # explicit coroutine for libraries that type check
    async def coroutine(self):
        await self

    def __await__(self):
        try:
            value = yield from self.payload().__await__()
        except Exception as err:
            self._result = None, err
        else:
            self._result = value, None
        self._done.set()

    def wait(self):
        self._done.wait()
        value, exception = self._result
        if exception is None:
            return value
        else:
            raise exception

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.payload)



from ..interfaces import Pool, CompositePool


class UniformComposite(CompositePool):
    """
    Uniform composition of several pools, with each pool weighted the same
    """
    children = []

    @property
    def demand(self):
        return self._demand

    @demand.setter
    def demand(self, value):
        self._demand = value
        child_count = len(self.children)
        for pool in self.children:
            pool.demand = value / child_count

    @property
    def supply(self):
        return sum(child.supply for child in self.children)

    @property
    def utilisation(self):
        try:
            return sum(child.utilisation for child in self.children) / len(self.children)
        except ZeroDivisionError:
            return 1.

    @property
    def allocation(self):
        try:
            return sum(child.allocation for child in self.children) / len(self.children)
        except ZeroDivisionError:
            return 1.

    def __init__(self, *children: Pool):
        self._demand = sum(child.demand for child in children)
        self.children = list(children)



from collections.abc import Mapping
from logging import Formatter, LogRecord

from .format_json import RECORD_ATTRIBUTES


def line_protocol(name, tags: dict = None, fields: dict = None, timestamp: float = None):
    """
    Format a report as per InfluxDB line protocol

    :param name: name of the report
    :param tags: tags identifying the specific report
    :param fields: measurements of the report
    :param timestamp: when the measurement was taken, in **seconds** since the epoch
    """
    output_str = name
    if tags:
        output_str += ','
        output_str += ','.join('%s=%s' % (key, value) for key, value in sorted(tags.items()))
    output_str += ' '
    output_str += ','.join(('%s=%r' % (key, value)).replace("'", '"') for key, value in sorted(fields.items()))
    if timestamp is not None:
        # line protocol requires nanosecond precision, python uses seconds
        output_str += ' %d' % (timestamp * 1E9)
    return output_str + '\n'


class LineProtocolFormatter(Formatter):
    """
    Formatter that emits data as InfluxDB Line Protocol

    :param tags: record data to use as tags
    :param resolution: resolution of timestamps in seconds

    The ``tags`` act as a whitelist for record keys if they are an iterable.
    When a dictionary is supplied, its values act as default values if the key is not in a record.
    """
    def __init__(self, tags: dict = None, resolution: float = None):
        super().__init__()
        self._default_tags = tags if isinstance(tags, Mapping) else {}
        self._tags_whitelist = set(tags) if tags is not None else set()
        self._fields_blacklist = self._tags_whitelist | set(RECORD_ATTRIBUTES)
        self._resolution = resolution

    @staticmethod
    def _as_dict(self, record, keys: set, defaults: dict):
        """Filter a populated ``record`` to a :py:class:`dict`"""
        data = defaults.copy()
        data.update({
            name: value
            for name, value in record.__dict__.items()
            if name in keys
        })
        return data

    def format(self, record: LogRecord):
        args = record.args
        assert isinstance(args, Mapping), 'monitor record argument must be a mapping, not %r' % type(args)
        record.asctime = self.formatTime(record, self.datefmt)
        record.message = record.getMessage()
        tags = self._default_tags.copy()
        tags.update({key: value for key, value in args.items() if key in self._tags_whitelist})
        fields = {key: value for key, value in args.items() if key not in self._fields_blacklist}
        timestamp = record.created // self._resolution * self._resolution if self._resolution is not None else None
        return line_protocol(name=record.message, tags=tags, fields=fields, timestamp=timestamp)


if __name__ == '__main__':
    import logging
    logger = logging.getLogger()
    logger.handlers = [logging.StreamHandler()]
    logger.handlers[0].formatter = LineProtocolFormatter({'latitude': 49, 'longitude': 8})
    logger.warning('forecast', {'temperature': 298, 'humidity': 0.45})



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

repo_base_dir = os.path.abspath(os.path.dirname(__file__))
# pull in the packages metadata
package_about = {}
with open(os.path.join(repo_base_dir, "cobald", "__about__.py")) as about_file:
    exec(about_file.read(), package_about)


with open(os.path.join(repo_base_dir, 'README.rst'), 'r') as README:
    long_description = README.read()

if __name__ == '__main__':
    setup(
        name=package_about['__title__'],
        version=package_about['__version__'],
        description=package_about['__summary__'],
        long_description=long_description.strip(),
        author=package_about['__author__'],
        author_email=package_about['__email__'],
        url=package_about['__url__'],
        packages=find_packages(),
        entry_points={
            'console_scripts': [
                'cobald = cobald.daemon.core:main',
            ],
        },
        # dependencies
        install_requires=[
            'typing',
            'pyyaml',
            'trio',
            'include',
        ],
        extras_require={
            'docs':  ["sphinx", "sphinxcontrib-tikz"],
        },
        # metadata for package search
        license='MIT',
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        classifiers=[
            'Development Status :: 2 - Pre-Alpha',
            'Intended Audience :: Developers',
            'Intended Audience :: Information Technology',
            'Intended Audience :: Science/Research',
            'Intended Audience :: System Administrators',
            'Topic :: Adaptive Technologies',
            'Topic :: Office/Business :: Scheduling',
            'Topic :: System :: Distributed Computing',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
        ],
        keywords=package_about['__keywords__'],
        # unit tests
        setup_requires=['pytest-runner'],
        test_suite='cobald_tests',
        tests_require=['pytest'],
    )



import trio

from cobald.interfaces import Pool, Controller

from cobald.daemon import service


@service(flavour=trio)
class LinearController(Controller):
    """
    Controller that linearly increases or decreases demand

    :param target: the pool to manage
    :param low_utilisation: pool utilisation below which resources are decreased
    :param high_allocation: pool allocation above which resources are increased
    :param rate: maximum change of demand in resources per second
    """
    @property
    def rate(self):
        return 1 / self._interval

    @rate.setter
    def rate(self, value):
        self._interval = 1 / value

    def __init__(self, target: Pool, low_utilisation=0.5, high_allocation=0.5, rate=1):
        super().__init__(target=target)
        self._interval = None
        assert rate > 0
        self.rate = rate
        assert low_utilisation <= high_allocation
        self.low_utilisation = low_utilisation
        self.high_allocation = high_allocation

    async def run(self):
        while True:
            self.regulate_demand()
            await trio.sleep(self._interval)

    def regulate_demand(self):
        if self.target.utilisation < self.low_utilisation:
            self.target.demand -= 1
        elif self.target.allocation > self.high_allocation:
            self.target.demand += 1



"""
Daemon core specific to cobald
"""
import os
import sys
import logging
import platform

import cobald

from .logger import initialise_logging
from .config.mapping import Translator
from .config.yaml import load_configuration as load_yaml_configuration
from .config.python import load_configuration as load_python_configuration
from .cli import CLI
from . import runtime
from .. import __about__


class PipelineTranslator(Translator):
    def translate_hierarchy(self, structure, *, where='', **construct_kwargs):
        try:
            pipeline = structure['pipeline']
        except (KeyError, TypeError):
            return super().translate_hierarchy(structure, where=where, **construct_kwargs)
        else:
            prev_item, items = None, []
            for index, item in reversed(list(enumerate(pipeline))):
                if prev_item is not None:
                    prev_item = self.translate_hierarchy(item, where='%s[%s]' % (where, index), target=prev_item)
                else:
                    prev_item = self.translate_hierarchy(item)
                items.append(prev_item)
            return list(reversed(items))


def core(configuration: str, level: str, target: str, short_format: bool):
    initialise_logging(level=level, target=target, short_format=short_format)
    logger = logging.getLogger(__package__)
    logger.info('COBalD %s', __about__.__version__)
    logger.info(__about__.__url__)
    logger.info('%s %s (%s)', platform.python_implementation(), platform.python_version(), sys.executable)
    logger.debug(cobald.__file__)
    if os.path.splitext(configuration)[1] == '.yaml':
        handle = load_yaml_configuration(configuration, translator=PipelineTranslator())
    elif os.path.splitext(configuration)[1] == '.py':
        handle = load_python_configuration(configuration)
    else:
        raise ValueError('Unknown configuration extension: %r' % os.path.splitext(configuration)[1])
    logger.info('Running main event loop...')
    runtime.accept()
    del handle


def main():
    options = CLI.parse_args()
    core(options.CONFIGURATION, options.log_level, options.log_target, options.log_journal)



import pytest

from ..mock.pool import MockPool

from cobald.controller.linear import LinearController


class TestLinearController(object):
    def test_init(self):
        with pytest.raises(TypeError):
            LinearController()

    def test_init_with_pool(self):
        pool = MockPool()
        controller = LinearController(target=pool)
        pool.utilisation = 1.0

        expected_demand = 0
        for i in range(2):
            assert(pool.demand == expected_demand)
            controller.regulate_demand()
            expected_demand += 1

        pool.utilisation = 0.0
        for i in range(2):
            assert (pool.demand == expected_demand)
            controller.regulate_demand()
            expected_demand -= 1

    def test_utilisation(self):
        pool = MockPool()
        controller = LinearController(target=pool, low_utilisation=0.5)

        old_demand = pool.demand = 1
        pool.utilisation = pool.allocation = 1.0
        controller.regulate_demand()
        assert(pool.demand - old_demand == 1)

        old_demand = pool.demand = 1
        pool.utilisation = pool.allocation = 0.0
        controller.regulate_demand()
        assert(pool.demand - old_demand == -1)

        old_demand = pool.demand = 1
        pool.utilisation = pool.allocation = controller.low_utilisation
        controller.regulate_demand()
        assert(pool.demand - old_demand == 0)

        old_demand = pool.demand = 1
        pool.utilisation = pool.allocation = controller.low_utilisation - .01
        controller.regulate_demand()
        assert (pool.demand - old_demand == -1)

        old_demand = pool.demand = 1
        pool.utilisation = pool.allocation = controller.low_utilisation + .01
        controller.regulate_demand()
        assert(pool.demand - old_demand == 1)

    def test_allocation(self):
        pool = MockPool()
        controller = LinearController(target=pool, high_allocation=0.5)

        old_demand = pool.demand = 1
        pool.allocation = pool.utilisation = 1.0
        controller.regulate_demand()
        assert(pool.demand - old_demand == 1)

        old_demand = pool.demand = 1
        pool.allocation = pool.utilisation = 0.0
        controller.regulate_demand()
        assert(pool.demand - old_demand == -1)

        old_demand = pool.demand = 1
        pool.allocation = pool.utilisation = controller.high_allocation
        controller.regulate_demand()
        assert(pool.demand - old_demand == 0)

        old_demand = pool.demand = 1
        pool.allocation = pool.utilisation = controller.high_allocation + 0.01
        controller.regulate_demand()
        assert (pool.demand - old_demand == 1)

        old_demand = pool.demand = 1
        pool.allocation = pool.utilisation = controller.high_allocation - 0.01
        controller.regulate_demand()
        assert (pool.demand - old_demand == -1)

    def test_low_high(self):
        pool = MockPool()
        with pytest.raises(AssertionError):
            LinearController(target=pool, low_utilisation=1, high_allocation=0)
        assert(LinearController(target=pool, low_utilisation=.5, high_allocation=.5))
        assert(LinearController(target=pool, low_utilisation=.1, high_allocation=.9))

    def test_rate(self):
        pool = MockPool()
        with pytest.raises(AssertionError):
            LinearController(target=pool, rate=0)
        controller = LinearController(target=pool, rate=1)
        assert controller.rate == 1
        controller = LinearController(target=pool, rate=2)
        assert controller.rate == 2
        controller = LinearController(target=pool, rate=1/3)
        assert controller.rate == 1/3




import logging
import threading

from ...utility.debug import NameRepr


class BaseRunner(object):
    flavour = None

    def __init__(self):
        self._logger = logging.getLogger('cobald.runtime.runner.%s' % NameRepr(self.flavour))
        self._payloads = []
        self._lock = threading.Lock()
        #: signal that runner should keep in running
        self.running = threading.Event()
        #: signal that runner has stopped
        self._stopped = threading.Event()
        self.running.clear()
        self._stopped.set()

    def __bool__(self):
        with self._lock:
            return bool(self._payloads) or self.running.is_set()

    def register_payload(self, payload):
        """
        Register ``payload`` for asynchronous execution

        This runs ``payload`` as an orphaned background task as soon as possible.
        It is an error for ``payload`` to return or raise anything without handling it.
        """
        with self._lock:
            self._payloads.append(payload)

    def run_payload(self, payload):
        """
        Register ``payload`` for synchronous execution

        This runs ``payload`` as soon as possible, blocking until completion.
        Should ``payload`` return or raise anything, it is propagated to the caller.
        """
        raise NotImplementedError

    def run(self):
        """
        Execute all current and future payloads

        Blocks and executes payloads until :py:meth:`stop` is called.
        It is an error for any orphaned payload to return or raise.
        """
        self._logger.info('runner started: %s', self)
        try:
            with self._lock:
                assert not self.running.is_set() and self._stopped.is_set(), 'cannot re-run: %s' % self
                self.running.set()
                self._stopped.clear()
            self._run()
        except Exception:
            self._logger.error('runner aborted: %s', self)
            raise
        else:
            self._logger.info('runner stopped: %s', self)
        finally:
            with self._lock:
                self.running.clear()
                self._stopped.set()

    def _run(self):
        raise NotImplementedError

    def stop(self):
        """Stop execution of all current and future payloads"""
        if not self.running.wait(0.2):
            return
        self._logger.debug('runner disabled: %s', self)
        with self._lock:
            self.running.clear()
        self._stopped.wait()


class OrphanedReturn(Exception):
    """A runnable returned a value without anyone to receive it"""
    def __init__(self, who, value):
        super().__init__('no caller to receive %s from %s' % (value, who))
        self.who = who
        self.value = value
